# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

use strict;
use warnings;

package Clownfish::CFC::Test::TestUtils;

our $VERSION = '0.006002';
$VERSION = eval $VERSION;

use Exporter 'import';
our @EXPORT_OK = qw( test_files_dir );

use File::Spec::Functions qw( catdir updir );

sub test_files_dir {
    my @dirs = (
        't',
        catdir( updir(), 'common', 'test' ),
    );

    for my $dir (@dirs) {
        my $cfbase = catdir( $dir, 'cfbase' );
        return $dir if -d $cfbase;
    }

    die("test files dir not found");
}

1;

